@@ -43,6 +43,7 @@ from django.core.urlresolvers import reverse
 from django.db import models
 from django.db.models.fields import FieldDoesNotExist
 from django.db.models.signals import post_delete
+from django.db.models.sql.aggregates import Aggregate as SQLAggregate
 from django.dispatch import receiver
 from django.template import Template, Context
 from django.template.defaultfilters import filesizeformat
@@ -1505,7 +1506,7 @@ class Image(models.Model):
             test_runs__test__test_id='lava',
             test_runs__attributes__name=self.build_number_attribute).extra(
             select={
-                'build_number': 'cast("dashboard_app_namedattribute"."value" as int)'
+                'build_number': 'convert_to_integer("dashboard_app_namedattribute"."value")',
                 }).extra(
             order_by=['-build_number'],
             )[:count]
@@ -1571,32 +1572,45 @@ class FilterMatch(object):
     TestRunFilter.get_test_runs.
     """
 
-    bundle = None
-    specific_results = None
-    result_count = None
-    pass_count = None
-    test_run = None
     filter = None
+    tag = None # either a date (bundle__uploaded_on) or a build number
+    test_runs = None
+    specific_results = None # Will stay none unless filter specifies a test case
+    pass_count = None # Only filled out for filters that dont specify a test
+    result_code = None # Ditto
+
+    def _format_test_result(self, result):
+        prefix = result.test_case.test.test_id + ':' + result.test_case.test_case_id + ' '
+        if result.test_case.units:
+            return prefix + '%s%s' % (result.measurement, result.units)
+        else:
+            return prefix + result.RESULT_MAP[result.result]
+
+    def _format_test_run(self, tr):
+        return "%s %s pass / %s total" % (
+            tr.test.test_id,
+            tr.denormalization.count_pass,
+            tr.denormalization.count_all())
+
+    def _format_many_test_runs(self):
+        return "%s pass / %s total" % (self.pass_count, self.result_count)
 
     def format_for_mail(self):
         r = [' ~%s/%s ' % (self.filter.owner.username, self.filter.name)]
-        if self.filter.test_case:
-            r.extend([
-                self.filter.test.test_id,
-                ':',
-                self.filter.test_case.test_case_id,
-                ])
-            for result in self.specific_results:
-                if self.filter.test_case.units:
-                    result_desc = '%s%s' % (result.measurement, result.units)
-                else:
-                    result_desc = result.RESULT_MAP[result.result]
-                r.extend([' ', result_desc])
-        elif self.filter.test:
-            r.append('%s %s pass/%s total' % (
-                self.filter.test.test_id, self.pass_count, self.result_count))
+        if not self.filter_data['tests']:
+            r.append(self._format_many_test_runs())
         else:
-            r.append('%s pass/%s total' % (self.pass_count, self.result_count))
+            for test in self.filter_data['tests']:
+                if not test.all_case_ids():
+                    for tr in self.test_runs:
+                        if tr.test == test.test:
+                            r.append('\n    ')
+                            r.append(self._format_test_run(tr))
+                for case_id in test.all_case_ids():
+                    for result in self.specific_results:
+                        if result.test_case.id == case_id:
+                            r.append('\n    ')
+                            r.append(self._format_test_result(result))
         r.append('\n')
         return ''.join(r)
 
@@ -1605,22 +1619,80 @@ class MatchMakingQuerySet(object):
     """Wrap a QuerySet and construct FilterMatchs from what the wrapped query
     set returns.
 
-    Just enough of the QuerySet API to work with DataTable."""
+    Just enough of the QuerySet API to work with DataTable (i.e. ordering and
+    slicing)."""
 
     model = TestRun
 
-    def __init__(self, queryset, filter):
+    def __init__(self, queryset, filter_data):
         self.queryset = queryset
-        self.filter = filter
+        self.filter_data = filter_data
+        if filter_data['build_number_attribute']:
+            self.key = 'build_number'
+            self.key_name = 'Build'
+        else:
+            self.key = 'bundle__uploaded_on'
+            self.key_name = 'Uploaded On'
 
     def _makeMatches(self, data):
-        raise NotImplementedError(self._makeMatches)
+        test_run_ids = set()
+        for datum in data:
+            test_run_ids.update(datum['id__arrayagg'])
+        r = []
+        trs = TestRun.objects.filter(id__in=test_run_ids).select_related(
+            'denormalization', 'bundle', 'bundle__bundle_stream', 'test')
+        trs_by_id = {}
+        for tr in trs:
+            trs_by_id[tr.id] = tr
+        case_ids = set()
+        for t in self.filter_data['tests']:
+            case_ids.update(t.all_case_ids())
+        if case_ids:
+            result_ids_by_tr_id = {}
+            results_by_tr_id = {}
+            values = TestResult.objects.filter(
+                test_case__id__in=case_ids,
+                test_run__id__in=test_run_ids).values_list(
+                'test_run__id', 'id')
+            result_ids = set()
+            for v in values:
+                result_ids_by_tr_id.setdefault(v[0], []).append(v[1])
+                result_ids.add(v[1])
+
+            results_by_id = {}
+            for result in TestResult.objects.filter(
+                id__in=list(result_ids)).select_related(
+                'test', 'test_case', 'test_run__bundle__bundle_stream'):
+                results_by_id[result.id] = result
+
+            for tr_id, result_ids in result_ids_by_tr_id.items():
+                rs = results_by_tr_id[tr_id] = []
+                for result_id in result_ids:
+                    rs.append(results_by_id[result_id])
+        for datum in data:
+            trs = []
+            for id in set(datum['id__arrayagg']):
+                trs.append(trs_by_id[id])
+            match = FilterMatch()
+            match.test_runs = trs
+            match.filter_data = self.filter_data
+            match.tag = datum[self.key]
+            if case_ids:
+                match.specific_results = []
+                for id in set(datum['id__arrayagg']):
+                    match.specific_results.extend(results_by_tr_id.get(id, []))
+            else:
+                match.pass_count = sum(tr.denormalization.count_pass for tr in trs)
+                match.result_count = sum(tr.denormalization.count_all() for tr in trs)
+            r.append(match)
+        return iter(r)
 
     def _wrap(self, queryset, **kw):
-        return self.__class__(queryset, self.filter, **kw)
+        return self.__class__(queryset, self.filter_data, **kw)
 
     def order_by(self, *args):
-        return self._wrap(self.queryset.order_by(*args))
+        # the generic tables code calls this even when it shouldn't...
+        return self
 
     def count(self):
         return self.queryset.count()
@@ -1633,102 +1705,6 @@ class MatchMakingQuerySet(object):
         return self._makeMatches(data)
 
 
-class SpecificTestCaseMatchMakingQuerySet(MatchMakingQuerySet):
-
-    def _makeMatches(self, runs):
-        results_by_run_id = {}
-        for run in runs:
-            results_by_run_id[run.id] = []
-        results = TestResult.objects.filter(
-            test_run_id__in=results_by_run_id.keys(),
-            test_case_id=self.filter.test_case.id)
-        for result in results:
-            results_by_run_id[result.test_run_id].append(result)
-        matches = []
-        for run in runs:
-            match = FilterMatch()
-            specific_results = results_by_run_id[result.test_run_id]
-            match.specific_results = specific_results
-            match.result_count = len(specific_results)
-            match.pass_count = len([r for r in specific_results if r.result == r.RESULT_PASS])
-            match.test_run = run
-            match.bundle = run.bundle
-            match.filter = self.filter
-            matches.append(match)
-        return iter(matches)
-
-
-
-class SpecificTestMatchMakingQuerySet(MatchMakingQuerySet):
-    def _makeMatches(self, runs):
-        matches = []
-        for run in runs:
-            match = FilterMatch()
-            match.specific_results = None
-            match.result_count = run.denormalization.count_all()
-            match.pass_count = run.denormalization.count_pass
-            match.test_run = run
-            match.bundle = run.bundle
-            match.filter = self.filter
-            matches.append(match)
-        return iter(matches)
-
-
-class BundleMatchMakingQuerySet(MatchMakingQuerySet):
-
-    model = Bundle
-
-    def __init__(self, queryset, filter, mis_ordered=False):
-        super(BundleMatchMakingQuerySet, self).__init__(queryset, filter)
-        self.mis_ordered = mis_ordered
-
-    def _makeMatches(self, bundles):
-        assert not self.mis_ordered, """
-           attempt to materialize BundleMatchMakingQuerySet when ordered on
-           non-bundle field"""
-        matches = []
-        counted_bundles = Bundle.objects.filter(
-            id__in=[b.id for b in bundles]).annotate(
-            pass_count=models.Sum('test_runs__denormalization__count_pass'),
-            unknown_count=models.Sum('test_runs__denormalization__count_unknown'),
-            skip_count=models.Sum('test_runs__denormalization__count_skip'),
-            fail_count=models.Sum('test_runs__denormalization__count_fail'))
-        bundles_by_id = {}
-        for bundle in counted_bundles:
-            bundles_by_id[bundle.id] = bundle
-        for bundle in bundles:
-            match = FilterMatch()
-            match.specific_results = None
-            cb = bundles_by_id[bundle.id]
-            match.result_count = cb.unknown_count + cb.skip_count + cb.pass_count + cb.fail_count
-            match.pass_count = cb.pass_count
-            match.test_run = None
-            match.bundle = bundle
-            match.filter = self.filter
-            matches.append(match)
-        return iter(matches)
-
-    def _wrap(self, queryset, **kw):
-        if 'mis_ordered' not in kw:
-            kw['mis_ordered'] = self.mis_ordered
-        return self.__class__(queryset, self.filter, **kw)
-
-    def order_by(self, field):
-        if field.startswith('bundle__') or field.startswith('-bundle__'):
-            if field.startswith('-'):
-                prefix = '-'
-                field = field[1:]
-            else:
-                prefix = ''
-            field = field[len('bundle__'):]
-            r = super(BundleMatchMakingQuerySet, self).order_by(
-                prefix+field)
-            r.mis_ordered = False
-            return r
-        else:
-            return self._wrap(self.queryset, mis_ordered=True)
-
-
 class TestRunFilterAttribute(models.Model):
 
     name = models.CharField(max_length=1024)
@@ -1740,6 +1716,50 @@ class TestRunFilterAttribute(models.Model):
         return '%s = %s' % (self.name, self.value)
 
 
+class TestRunFilterTest(models.Model):
+
+    test = models.ForeignKey(Test, related_name="+")
+    filter = models.ForeignKey("TestRunFilter", related_name="tests")
+    index = models.PositiveIntegerField(
+        help_text = _(u"The index of this test in the filter"))
+
+    def all_case_ids(self):
+        return self.cases.all().order_by('index').values_list('test_case__id', flat=True)
+
+    def all_case_names(self):
+        return self.cases.all().order_by('index').values_list('test_case__test_case_id', flat=True)
+
+    def __unicode__(self):
+        return unicode(self.test)
+
+
+class TestRunFilterTestCase(models.Model):
+
+    test_case = models.ForeignKey(TestCase, related_name="+")
+    test = models.ForeignKey(TestRunFilterTest, related_name="cases")
+    index = models.PositiveIntegerField(
+        help_text = _(u"The index of this case in the test"))
+
+    def __unicode__(self):
+        return unicode(self.test_case)
+
+
+class SQLArrayAgg(SQLAggregate):
+    sql_function = 'array_agg'
+
+
+class ArrayAgg(models.Aggregate):
+    name = 'ArrayAgg'
+    def add_to_query(self, query, alias, col, source, is_summary):
+        aggregate = SQLArrayAgg(
+            col, source=source, is_summary=is_summary, **self.extra)
+        # For way more detail than you want about what this next line is for,
+        # see
+        # http://voices.canonical.com/michael.hudson/2012/09/02/using-postgres-array_agg-from-django/
+        aggregate.field = models.DecimalField() # vomit
+        query.aggregates[alias] = aggregate
+
+
 class TestRunFilter(models.Model):
 
     owner = models.ForeignKey(User)
@@ -1755,41 +1775,24 @@ class TestRunFilter(models.Model):
     bundle_streams = models.ManyToManyField(BundleStream)
     bundle_streams.help_text = 'A filter only matches tests within the given <b>bundle streams</b>.'
 
-    test = models.ForeignKey(
-        Test, blank=True, null=True,
-        help_text=("A filter can optionally be restricted to a particular "
-                   "<b>test</b>, or even a <b>test case</b> within a test."))
-
-    test_case = models.ForeignKey(TestCase, blank=True, null=True)
-
     public = models.BooleanField(
         default=False, help_text="Whether other users can see this filter.")
 
+    build_number_attribute = models.CharField(
+        max_length=1024, blank=True, null=True,
+        help_text="For some filters, there is a natural <b>build number</b>.  If you specify the name of the attribute that contains the build number here, the results of the filter will be grouped and ordered by this build number.")
+
     @property
     def summary_data(self):
         return {
             'bundle_streams': self.bundle_streams.all(),
             'attributes': self.attributes.all().values_list('name', 'value'),
-            'test': self.test,
-            'test_case': self.test_case,
+            'tests': self.tests.all().prefetch_related('cases'),
+            'build_number_attribute': self.build_number_attribute,
             }
 
     def __unicode__(self):
-        test = self.test
-        if not test:
-            test = "<any>"
-        test_case = self.test_case
-        if not test_case:
-            test_case = "<any>"
-        attrs = []
-        for attr in self.attributes.all():
-            attrs.append(unicode(attr))
-        attrs = ', '.join(attrs)
-        if attrs:
-            attrs = ' ' + attrs + '; '
-        return "<TestRunFilter ~%s/%s %d streams;%s %s:%s>" % (
-            self.owner.username, self.name, self.bundle_streams.count(), attrs, test, test_case)
-
+        return "<TestRunFilter ~%s/%s>" % (self.owner.username, self.name)
 
     # given filter:
     # select from testrun
@@ -1798,41 +1801,63 @@ class TestRunFilter(models.Model):
     #    and testrun has attribute with key = key2 and value = value2
     #    and               ...
     #    and testrun has attribute with key = keyN and value = valueN
-    #    and testrun has filter.test/testcase requested
+    #    and testrun has any of the tests/testcases requested
 
-    def get_test_runs_impl(self, user, bundle_streams, attributes):
+    def get_test_runs_impl(self, user, bundle_streams, attributes, tests):
         accessible_bundle_streams = BundleStream.objects.accessible_by_principal(
             user)
-        testruns = TestRun.objects.filter(
-            models.Q(bundle__bundle_stream__in=accessible_bundle_streams),
-            models.Q(bundle__bundle_stream__in=bundle_streams),
-            )
+        bs_ids = [bs.id for bs in set(accessible_bundle_streams) & set(bundle_streams)]
+        conditions = [models.Q(bundle__bundle_stream__id__in=bs_ids)]
+
+        content_type_id = ContentType.objects.get_for_model(TestRun).id
 
         for (name, value) in attributes:
-            testruns = TestRun.objects.filter(
-                id__in=testruns.values_list('id'),
-                attributes__name=name, attributes__value=value)
+            # We punch through the generic relation abstraction here for 100x
+            # better performance.
+            conditions.append(
+                models.Q(id__in=NamedAttribute.objects.filter(
+                    name=name, value=value, content_type_id=content_type_id
+                    ).values('object_id')))
 
-        if self.test_case:
-            testruns = TestRun.objects.filter(
-                id__in=testruns.values_list('id'),
-                test_results__test_case=self.test_case,
-                test=self.test_case.test)
-            wrapper_cls = SpecificTestCaseMatchMakingQuerySet
-        elif self.test:
-            testruns = TestRun.objects.filter(
-                id__in=testruns.values_list('id'),
-                test=self.test)
-            wrapper_cls = SpecificTestMatchMakingQuerySet
+        test_condition = None
+        for test in tests:
+            cases = list(test.all_case_ids())
+            if cases:
+                q = models.Q(
+                    test__id=test.test.id,
+                    test_results__test_case__id__in=cases)
+            else:
+                q = models.Q(test__id=test.test.id)
+            if test_condition:
+                test_condition = test_condition | q
+            else:
+                test_condition = q
+        if test_condition:
+            conditions.append(test_condition)
+
+        testruns = TestRun.objects.filter(*conditions)
+
+        if self.build_number_attribute:
+            testruns = testruns.filter(
+                attributes__name=self.build_number_attribute).extra(
+                select={
+                    'build_number': 'convert_to_integer("dashboard_app_namedattribute"."value")',
+                    },
+                where=['convert_to_integer("dashboard_app_namedattribute"."value") IS NOT NULL']).extra(
+                order_by=['-build_number'],
+                ).values('build_number').annotate(ArrayAgg('id'))
         else:
-            # if the filter doesn't specify a test, we still only return one
-            # test run per bundle.  the display code knows to do different
-            # things in this case.
-            testruns = Bundle.objects.filter(
-                test_runs__in=testruns)
-            wrapper_cls = BundleMatchMakingQuerySet
+            testruns = testruns.order_by('-bundle__uploaded_on').values(
+                'bundle__uploaded_on').annotate(ArrayAgg('id'))
 
-        return wrapper_cls(testruns, self)
+        filter_data = {
+            'bundle_streams': bundle_streams,
+            'attributes': attributes,
+            'tests': tests,
+            'build_number_attribute': self.build_number_attribute,
+            }
+
+        return MatchMakingQuerySet(testruns, filter_data)
 
     # given bundle:
     # select from filter
@@ -1845,15 +1870,8 @@ class TestRunFilter(models.Model):
 
     @classmethod
     def matches_against_bundle(self, bundle):
-        filters = bundle.bundle_stream.testrunfilter_set.all()
-        filters = filters.filter(
-            models.Q(test__isnull=True)
-            |models.Q(test__in=bundle.test_runs.all().values('test')))
-        filters = filters.filter(
-            models.Q(test_case__isnull=True)
-            |models.Q(test_case__in=TestResult.objects.filter(
-                test_run__in=bundle.test_runs.all()).values('test_case')))
-        filters = filters.extra(
+        bundle_filters = bundle.bundle_stream.testrunfilter_set.all()
+        attribute_filters = list(bundle_filters.extra(
             where=[
             """(select min((select count(*)
                               from dashboard_app_testrunfilterattribute
@@ -1864,41 +1882,43 @@ class TestRunFilter(models.Model):
                                           select django_content_type.id from django_content_type
                                           where app_label = 'dashboard_app' and model='testrun')
                                  and object_id = dashboard_app_testrun.id)))
-            from dashboard_app_testrun where dashboard_app_testrun.bundle_id = %s) = 0 """ % bundle.id],
+            from dashboard_app_testrun where dashboard_app_testrun.bundle_id = %s) = 0""" % bundle.id],
+            ))
+        no_test_filters = []#list(attribute_filters.annotate(models.Count('tests')).filter(tests__count=0))
+        no_test_case_filters = list(
+            TestRunFilter.objects.filter(
+                id__in=TestRunFilterTest.objects.filter(
+                    filter__in=attribute_filters, test__in=bundle.test_runs.all().values('test_id')).annotate(
+                    models.Count('cases')).filter(cases__count=0).values('filter__id'),
+                ))
+        tcf = TestRunFilter.objects.filter(
+            id__in=TestRunFilterTest.objects.filter(
+                filter__in=attribute_filters,
+                cases__test_case__id__in=bundle.test_runs.all().values('test_results__test_case__id')
+                ).values('filter__id')
             )
-        filters = list(filters)
+        test_case_filters = list(tcf)
+
+        filters = set(test_case_filters + no_test_case_filters + no_test_filters)
         matches = []
+        bundle_with_counts = Bundle.objects.annotate(
+            pass_count=models.Sum('test_runs__denormalization__count_pass'),
+            unknown_count=models.Sum('test_runs__denormalization__count_unknown'),
+            skip_count=models.Sum('test_runs__denormalization__count_skip'),
+            fail_count=models.Sum('test_runs__denormalization__count_fail')).get(
+            id=bundle.id)
         for filter in filters:
-            if filter.test:
-                for test_run in bundle.test_runs.filter(test=filter.test):
-                    match = FilterMatch()
-                    match.filter = filter
-                    match.test_run = test_run
-                    if filter.test_case:
-                        match.specific_results = list(
-                            test_run.test_results.filter(test_case=filter.test_case))
-                        match.result_count = len(match.specific_results)
-                        match.pass_count = len(
-                            [r for r in match.specific_results if r.result == r.RESULT_PASS])
-                    else:
-                        match.specific_results = None
-                        match.result_count = test_run.denormalization.count_all()
-                        match.pass_count = test_run.denormalization.count_pass
-                    matches.append(match)
-            else:
-                match = FilterMatch()
-                match.filter = filter
-                match.test_run = None
-                bundle_with_counts = Bundle.objects.annotate(
-                    pass_count=models.Sum('test_runs__denormalization__count_pass'),
-                    unknown_count=models.Sum('test_runs__denormalization__count_unknown'),
-                    skip_count=models.Sum('test_runs__denormalization__count_skip'),
-                    fail_count=models.Sum('test_runs__denormalization__count_fail')).get(
-                    id=bundle.id)
-                match.specific_results = None
-                b = bundle_with_counts
-                match.result_count = b.unknown_count + b.skip_count + b.pass_count + b.fail_count
-                match.pass_count = bundle_with_counts.pass_count
+            match = FilterMatch()
+            match.filter = filter
+            match.filter_data = filter.summary_data
+            match.test_runs = list(bundle.test_runs.all())
+            match.specific_results = list(
+                TestResult.objects.filter(
+                    test_case__id__in=filter.tests.all().values('cases__test_case__id'),
+                    test_run__bundle=bundle))
+            b = bundle_with_counts
+            match.result_count = b.unknown_count + b.skip_count + b.pass_count + b.fail_count
+            match.pass_count = bundle_with_counts.pass_count
             matches.append(match)
         return matches
 
@@ -1906,7 +1926,8 @@ class TestRunFilter(models.Model):
         return self.get_test_runs_impl(
             user,
             self.bundle_streams.all(),
-            self.attributes.values_list('name', 'value'))
+            self.attributes.values_list('name', 'value'),
+            self.tests.all())
 
     @models.permalink
     def get_absolute_url(self):
@@ -1953,8 +1974,27 @@ class TestRunFilterSubscription(models.Model):
         recipients = {}
         for sub in subscriptions:
             match = matches_by_filter_id[sub.filter.id]
-            if sub.level == cls.NOTIFICATION_FAILURE and match.pass_count == match.result_count:
-                continue
+            if sub.level == cls.NOTIFICATION_FAILURE:
+                failure_found = False
+                if not match.filter_data['tests']:
+                    failure_found = match.pass_count != match.result_count
+                else:
+                    for t in match.filter_data['tests']:
+                        if not t.all_case_ids():
+                            for tr in match.test_runs:
+                                if tr.test == t.test:
+                                    if tr.denormalization.count_pass != tr.denormalization.count_all():
+                                        failure_found = True
+                                        break
+                        if failure_found:
+                            break
+                if not failure_found:
+                    for r in match.specific_results:
+                        if r.result != TestResult.RESULT_PASS:
+                            failure_found = True
+                            break
+                if not failure_found:
+                    continue
             recipients.setdefault(sub.user, []).append(match)
         return recipients
 
