@@ -20,18 +20,22 @@
 Views for the Dashboard application
 """
 
+import operator
 import re
 import json
 
 from django.conf import settings
 from django.contrib.admin.widgets import FilteredSelectMultiple
 from django.contrib.auth.decorators import login_required
+from django.contrib.contenttypes.models import ContentType
 from django.contrib.sites.models import Site
 from django.core.exceptions import PermissionDenied, ValidationError
 from django.core.urlresolvers import reverse
 from django.db.models.manager import Manager
 from django.db.models.query import QuerySet
 from django import forms
+from django.forms.formsets import BaseFormSet, formset_factory
+from django.forms.widgets import Select
 from django.http import Http404, HttpResponse, HttpResponseRedirect
 from django.shortcuts import render_to_response, redirect, get_object_or_404
 from django.template import RequestContext, loader
@@ -456,6 +460,12 @@ class UserFiltersTable(DataTablesTable):
     {% endfor %}
     ''')
 
+    build_number_attribute = Column()
+    def render_build_number_attribute(self, value):
+        if not value:
+            return ''
+        return value
+
     attributes = TemplateColumn('''
     {% for a in record.attributes.all %}
     {{ a }}  <br />
@@ -463,13 +473,24 @@ class UserFiltersTable(DataTablesTable):
     ''')
 
     test = TemplateColumn('''
-    {% if record.test_case %}
-        {{ record.test }}:{{ record.test_case }}
-    {% elif record.test %}
-        {{ record.test }}:&lt;any&gt;
-    {% else %}
-        &lt;any&gt;:&lt;any&gt;
-    {% endif %}
+      <table style="border-collapse: collapse">
+        <tbody>
+          {% for test in record.tests.all %}
+          <tr>
+            <td>
+              {{ test.test }}
+            </td>
+            <td>
+              {% for test_case in test.all_case_names %}
+              {{ test_case }}
+              {% empty %}
+              <i>any</i>
+              {% endfor %}
+            </td>
+          </tr>
+          {% endfor %}
+        </tbody>
+      </table>
     ''')
 
     subscription = Column()
@@ -523,10 +544,27 @@ def filters_list(request):
     )
 
 
+class TestRunColumn(Column):
+    def render(self, record):
+        # This column is only rendered if we don't really expect
+        # record.test_runs to be very long...
+        links = []
+        trs = [tr for tr in record.test_runs if tr.test.test_id == self.verbose_name]
+        for tr in trs:
+            text = '%s / %s' % (tr.denormalization.count_pass, tr.denormalization.count_all())
+            links.append('<a href="%s">%s</a>' % (tr.get_absolute_url(), text))
+        return mark_safe('&nbsp;'.join(links))
+
+
 class SpecificCaseColumn(Column):
-    def render(self, value, record):
+    def __init__(self, verbose_name, test_case_id):
+        super(SpecificCaseColumn, self).__init__(verbose_name)
+        self.test_case_id = test_case_id
+    def render(self, record):
         r = []
-        for result in value:
+        for result in record.specific_results:
+            if result.test_case_id != self.test_case_id:
+                continue
             if result.result == result.RESULT_PASS and result.units:
                 s = '%s %s' % (result.measurement, result.units)
             else:
@@ -542,53 +580,74 @@ class BundleColumn(Column):
 
 class FilterTable(DataTablesTable):
     def __init__(self, *args, **kwargs):
-        filter = kwargs['params'][1]
-        data = filter.summary_data
+        kwargs['template'] = 'dashboard_app/filter_results_table.html'
         super(FilterTable, self).__init__(*args, **kwargs)
-        if len(data['bundle_streams']) == 1:
-            del self.base_columns['bundle_stream']
-        if data['test_case']:
-            del self.base_columns['bundle']
+        match_maker = self.data.queryset
+        self.base_columns['tag'].verbose_name = match_maker.key_name
+        bundle_stream_col = self.base_columns.pop('bundle_stream')
+        bundle_col = self.base_columns.pop('bundle')
+        tag_col = self.base_columns.pop('tag')
+        self.complex_header = False
+        if match_maker.filter_data['tests']:
             del self.base_columns['passes']
             del self.base_columns['total']
-            self.base_columns['specific_results'].verbose_name = mark_safe(
-                data['test_case'].test_case_id)
-        elif data['test']:
-            del self.base_columns['bundle']
-            del self.base_columns['specific_results']
+            for i, t in enumerate(reversed(match_maker.filter_data['tests'])):
+                if len(t.all_case_names()) == 0:
+                    col = TestRunColumn(mark_safe(t.test.test_id))
+                    self.base_columns.insert(0, 'test_run_%s' % i, col)
+                elif len(t.all_case_names()) == 1:
+                    n = t.test.test_id + ':' + t.all_case_names()[0]
+                    col = SpecificCaseColumn(mark_safe(n), t.all_case_ids()[0])
+                    self.base_columns.insert(0, 'test_run_%s_case' % i, col)
+                else:
+                    col0 = SpecificCaseColumn(mark_safe(t.all_case_names()[0]), t.all_case_ids()[0])
+                    col0.in_group = True
+                    col0.first_in_group = True
+                    col0.group_length = len(t.all_case_names())
+                    col0.group_name = mark_safe(t.test.test_id)
+                    self.complex_header = True
+                    self.base_columns.insert(0, 'test_run_%s_case_%s' % (i, 0), col0)
+                    for j, n in enumerate(t.all_case_names()[1:], 1):
+                        col = SpecificCaseColumn(mark_safe(n), t.all_case_ids()[j])
+                        col.in_group = True
+                        col.first_in_group = False
+                        self.base_columns.insert(j, 'test_run_%s_case_%s' % (i, j), col)
         else:
-            del self.base_columns['test_run']
-            self.base_columns['passes']
-            self.base_columns['total']
-            del self.base_columns['specific_results']
-        uploaded_col_index = self.base_columns.keys().index('uploaded_on')
-        self.datatable_opts = self.datatable_opts.copy()
-        self.datatable_opts['aaSorting'] = [[uploaded_col_index, 'desc']]
-        self._compute_queryset(kwargs['params'])
+            self.base_columns.insert(0, 'bundle', bundle_col)
+        if len(match_maker.filter_data['bundle_streams']) > 1:
+            self.base_columns.insert(0, 'bundle_stream', bundle_stream_col)
+        self.base_columns.insert(0, 'tag', tag_col)
 
-    bundle_stream = Column(accessor='bundle.bundle_stream')
+    tag = Column()
 
-    bundle = BundleColumn(accessor='bundle', sortable=False)
+    def render_bundle_stream(self, record):
+        bundle_streams = set(tr.bundle.bundle_stream for tr in record.test_runs)
+        links = []
+        for bs in sorted(bundle_streams, key=operator.attrgetter('pathname')):
+            links.append('<a href="%s">%s</a>' % (
+                bs.get_absolute_url(), escape(bs.pathname)))
+        return mark_safe('<br />'.join(links))
+    bundle_stream = Column(mark_safe("Bundle Stream(s)"))
 
-    test_run = TemplateColumn(
-        '<a href="{{ record.test_run.get_absolute_url }}">'
-        '<code>{{ record.test_run.test }} results<code/></a>',
-        accessor="test__test_id",
-        )
+    def render_bundle(self, record):
+        bundles = set(tr.bundle for tr in record.test_runs)
+        links = []
+        for b in sorted(bundles, key=operator.attrgetter('uploaded_on')):
+            links.append('<a href="%s">%s</a>' % (
+                b.get_absolute_url(), escape(b.content_filename)))
+        return mark_safe('<br />'.join(links))
+    bundle = Column(mark_safe("Bundle(s)"))
 
-    uploaded_on = TemplateColumn(
-        '{{ record.bundle.uploaded_on|date:"Y-m-d H:i:s" }}',
-        accessor='bundle__uploaded_on')
+    passes = Column(accessor='pass_count')
+    total = Column(accessor='result_count')
 
-    passes = Column(accessor='pass_count', sortable=False)
-    total = Column(accessor='result_count', sortable=False)
-    specific_results = SpecificCaseColumn(accessor='specific_results', sortable=False)
     def get_queryset(self, user, filter):
         return filter.get_test_runs(user)
 
     datatable_opts = {
         "sPaginationType": "full_numbers",
         "iDisplayLength": 25,
+        "bSort": False,
         }
 
 
@@ -699,10 +758,108 @@ var test_case_url = "{% url dashboard_app.views.filter_add_cases_for_test_json %
 var attr_name_completion_url = "{% url dashboard_app.views.filter_attr_name_completion_json %}";
 var attr_value_completion_url = "{% url dashboard_app.views.filter_attr_value_completion_json %}";
 </script>
+<script type="text/javascript" src="{{ STATIC_URL }}dashboard_app/js/jquery.formset.js"></script>
 <script type="text/javascript" src="{{ STATIC_URL }}dashboard_app/js/filter-edit.js"></script>
 '''
 
 
+class AttributesForm(forms.Form):
+    name = forms.CharField(max_length=1024)
+    value = forms.CharField(max_length=1024)
+
+AttributesFormSet = formset_factory(AttributesForm, extra=0)
+
+
+
+class TruncatingSelect(Select):
+
+    def render_option(self, selected_choices, option_value, option_label):
+        if len(option_label) > 50:
+            option_label = option_label[:50] + '...'
+        return super(TruncatingSelect, self).render_option(
+            selected_choices, option_value, option_label)
+
+
+class TRFTestCaseForm(forms.Form):
+
+    test_case = forms.ModelChoiceField(
+        queryset=TestCase.objects.none(), widget=TruncatingSelect, empty_label=None)
+
+
+class BaseTRFTestCaseFormSet(BaseFormSet):
+
+    def __init__(self, *args, **kw):
+        self._queryset = kw.pop('queryset')
+        super(BaseTRFTestCaseFormSet, self).__init__(*args, **kw)
+
+    def add_fields(self, form, index):
+        super(BaseTRFTestCaseFormSet, self).add_fields(form, index)
+        if self._queryset is not None:
+            form.fields['test_case'].queryset = self._queryset
+
+
+TRFTestCaseFormSet = formset_factory(
+    TRFTestCaseForm, extra=0, formset=BaseTRFTestCaseFormSet)
+
+
+class TRFTestForm(forms.Form):
+
+    def __init__(self, *args, **kw):
+        super(TRFTestForm, self).__init__(*args, **kw)
+        kw['initial'] = kw.get('initial', {}).get('test_cases', None)
+        kw.pop('empty_permitted', None)
+        kw['queryset'] = None
+        v = self['test'].value()
+        if v:
+            test = self.fields['test'].to_python(v)
+            queryset = TestCase.objects.filter(test=test).order_by('test_case_id')
+            kw['queryset'] = queryset
+        self.test_case_formset = TRFTestCaseFormSet(*args, **kw)
+
+    def is_valid(self):
+        return super(TRFTestForm, self).is_valid() and \
+               self.test_case_formset.is_valid()
+
+    def full_clean(self):
+        super(TRFTestForm, self).full_clean()
+        self.test_case_formset.full_clean()
+
+    test = forms.ModelChoiceField(
+        queryset=Test.objects.order_by('test_id'), required=True)
+
+
+class BaseTRFTestsFormSet(BaseFormSet):
+
+    def is_valid(self):
+        if not super(BaseTRFTestsFormSet, self).is_valid():
+            return False
+        for form in self.forms:
+            if not form.is_valid():
+                return False
+        return True
+
+
+TRFTestsFormSet = formset_factory(
+    TRFTestForm, extra=0, formset=BaseTRFTestsFormSet)
+
+
+class FakeTRFTest(object):
+    def __init__(self, form):
+        self.test = form.cleaned_data['test']
+        self.test_id = self.test.id
+        self._case_ids = []
+        self._case_names = []
+        for tc_form in form.test_case_formset:
+            self._case_ids.append(tc_form.cleaned_data['test_case'].id)
+            self._case_names.append(tc_form.cleaned_data['test_case'].test_case_id)
+
+    def all_case_ids(self):
+        return self._case_ids
+
+    def all_case_names(self):
+        return self._case_names
+
+
 class TestRunFilterForm(forms.ModelForm):
     class Meta:
         model = TestRunFilter
@@ -718,12 +875,6 @@ class TestRunFilterForm(forms.ModelForm):
             Context({'STATIC_URL': settings.STATIC_URL})
             )) + super_media
 
-    test = forms.ModelChoiceField(
-        queryset=Test.objects.order_by('test_id'), empty_label="<any>", required=False)
-
-    test_case = forms.ModelChoiceField(
-        queryset=TestCase.objects.none(), empty_label="<any>", required=False)
-
     def validate_name(self, value):
         self.instance.name = value
         try:
@@ -739,49 +890,79 @@ class TestRunFilterForm(forms.ModelForm):
         instance = super(TestRunFilterForm, self).save(commit=commit, **kwargs)
         if commit:
             instance.attributes.all().delete()
-            for (name, value) in self.attributes:
-                instance.attributes.create(name=name, value=value)
+            for a in self.attributes_formset.cleaned_data:
+                instance.attributes.create(name=a['name'], value=a['value'])
+            instance.tests.all().delete()
+            for i, test_form in enumerate(self.tests_formset.forms):
+                trf_test = instance.tests.create(
+                    test=test_form.cleaned_data['test'], index=i)
+                for j, test_case_form in enumerate(test_form.test_case_formset.forms):
+                    trf_test.cases.create(
+                        test_case=test_case_form.cleaned_data['test_case'], index=j)
         return instance
 
+    def is_valid(self):
+        return super(TestRunFilterForm, self).is_valid() and \
+               self.attributes_formset.is_valid() and \
+               self.tests_formset.is_valid()
+
+    def full_clean(self):
+        super(TestRunFilterForm, self).full_clean()
+        self.attributes_formset.full_clean()
+        self.tests_formset.full_clean()
+
     @property
     def summary_data(self):
         data = self.cleaned_data.copy()
-        data['attributes'] = self.attributes
+        tests = []
+        for form in self.tests_formset.forms:
+            tests.append(FakeTRFTest(form))
+        data['attributes'] = [
+            (d['name'], d['value']) for d in self.attributes_formset.cleaned_data]
+        data['tests'] = tests
         return data
 
     def __init__(self, user, *args, **kwargs):
         super(TestRunFilterForm, self).__init__(*args, **kwargs)
         self.instance.owner = user
-        self.fields['bundle_streams'].queryset = BundleStream.objects.accessible_by_principal(user)
+        kwargs.pop('instance', None)
+
+        attr_set_args = kwargs.copy()
+        if self.instance.pk:
+            initial = []
+            for attr in self.instance.attributes.all():
+                initial.append({
+                    'name': attr.name,
+                    'value': attr.value,
+                    })
+            attr_set_args['initial'] = initial
+        attr_set_args['prefix'] = 'attributes'
+        self.attributes_formset = AttributesFormSet(*args, **attr_set_args)
+
+        tests_set_args = kwargs.copy()
+        if self.instance.pk:
+            initial = []
+            for test in self.instance.tests.all().order_by('index').prefetch_related('cases'):
+                initial.append({
+                    'test': test.test,
+                    'test_cases': [{'test_case': unicode(tc.test_case.id)} for tc in test.cases.all().order_by('index')],
+                    })
+            tests_set_args['initial'] = initial
+        tests_set_args['prefix'] = 'tests'
+        self.tests_formset = TRFTestsFormSet(*args, **tests_set_args)
+
+        self.fields['bundle_streams'].queryset = \
+            BundleStream.objects.accessible_by_principal(user).order_by('pathname')
         self.fields['name'].validators.append(self.validate_name)
-        test = self['test'].value()
-        if test:
-            if not isinstance(test, int):
-                test = int(repr(test)[2:-1])
-            test = Test.objects.get(pk=test)
-            self.fields['test_case'].queryset = TestCase.objects.filter(test=test).order_by('test_case_id')
-
-    @property
-    def attributes(self):
-        if not self.is_bound and self.instance.pk:
-            return self.instance.attributes.values_list('name', 'value')
-        else:
-            attributes = []
-            for (var, value) in self.data.iteritems():
-                if var.startswith('attribute_key_'):
-                    index = int(var[len('attribute_key_'):])
-                    attr_value = self.data['attribute_value_' + str(index)]
-                    attributes.append((index, value, attr_value))
-
-            attributes.sort()
-            attributes = [a[1:] for a in attributes]
-            return attributes
 
     def get_test_runs(self, user):
         assert self.is_valid(), self.errors
         filter = self.save(commit=False)
+        tests = []
+        for form in self.tests_formset.forms:
+            tests.append(FakeTRFTest(form))
         return filter.get_test_runs_impl(
-            user, self.cleaned_data['bundle_streams'], self.attributes)
+            user, self.cleaned_data['bundle_streams'], self.summary_data['attributes'], tests)
 
 
 def filter_form(request, bread_crumb_trail, instance=None):
@@ -806,6 +987,7 @@ def filter_form(request, bread_crumb_trail, instance=None):
                     }, RequestContext(request))
     else:
         form = TestRunFilterForm(request.user, instance=instance)
+
     return render_to_response(
         'dashboard_app/filter_add.html', {
             'bread_crumb_trail': bread_crumb_trail,
@@ -859,8 +1041,10 @@ def filter_add_cases_for_test_json(request):
 
 def filter_attr_name_completion_json(request):
     term = request.GET['term']
+    content_type_id = ContentType.objects.get_for_model(TestRun).id
     result = NamedAttribute.objects.filter(
-        name__startswith=term).distinct('name').order_by('name').values_list('name', flat=True)
+        name__startswith=term, content_type_id=content_type_id
+        ).distinct().order_by('name').values_list('name', flat=True)
     return HttpResponse(
         json.dumps(list(result)),
         mimetype='application/json')
@@ -869,9 +1053,10 @@ def filter_attr_name_completion_json(request):
 def filter_attr_value_completion_json(request):
     name = request.GET['name']
     term = request.GET['term']
+    content_type_id = ContentType.objects.get_for_model(TestRun).id
     result = NamedAttribute.objects.filter(
-        name=name,
-        value__startswith=term).distinct('value').order_by('value').values_list('value', flat=True)
+        name=name, content_type_id=content_type_id, value__startswith=term
+        ).distinct().order_by('value').values_list('value', flat=True)
     return HttpResponse(
         json.dumps(list(result)),
         mimetype='application/json')
